import argparse
import copy
import functools
from inspect import isclass, getfullargspec
from typing import Callable, Dict, Type, Any

import yaml

from .node import ConfigFlexNode, ConfigNode, ConfigNodeProps, ConfigParam
from .utils import type_evaluator, read_raw_file


def config_to_dict(config_node: ConfigNode, *args, **kwargs):
    return ConfigNodeProps.get_props(config_node).to_dict(*args, **kwargs)


def config_to_yaml(config_node: ConfigNode, *args, **kwargs):
    return ConfigNodeProps.get_props(config_node).to_yaml(*args, **kwargs)


def config_from_string(config_node: Type[ConfigNode], string: str) -> ConfigNode:
    return config_node(yaml.safe_load(string))


def config_from_file(config_node: Type[ConfigNode], file_path: str) -> ConfigNode:
    content = read_raw_file(file_path)
    return config_from_file(config_node=config_node, string=content)


class config_from_dict:
    def __new__(cls,
                config: Dict[str, Any],
                annotations: Dict[str, Any] = None,
                name: str = None,
                flex: bool = False):
        return cls.cls(config=config, annotations=annotations, name=name, flex=flex)()

    @classmethod
    def cls(
        cls,
        config: Dict[str, Any],
        annotations: Dict[str, Any] = None,
        name: str = None,
        flex: bool = False
    ) -> Type[ConfigNode]:

        # set to empty dictionary if not provided
        annotations = annotations or {}

        # we need a name, does't have to mean anything and it
        # can be repeated, so we use the mem location of config
        # and annotations here.
        name = name or f'{id(config)}_{id(annotations)}'

        # Small function to let us look up a suitable type for a
        # parameter by either looking at type annotations, or by
        # checking the defaults; if anything fails; return a no-op
        # lambda as pseudo-default.
        def _get_ann(p_name: str) -> Callable:
            if p_name in annotations:
                return annotations[p_name]
            if p_name in config:
                return type(config[p_name])
            else:
                return lambda p_value: p_value

        # setting up type annotations, including for target
        param_annotations = {p: ConfigParam(_get_ann(p)) for p in
                             {*annotations, *config}}

        # get node class and set up defaults and annotations
        node_cls = type(name, (ConfigFlexNode if flex else ConfigNode, ), {})
        [setattr(node_cls, p, d) for p, d in config.items()]
        setattr(node_cls, '__annotations__', param_annotations)

        return node_cls


class config_to_program:
    @classmethod
    def _check_signature(cls, fn):
        expected_args = getfullargspec(fn).args
        if len(expected_args) == 0:
            msg = (f'Function `{fn.__name__}` cannot be decorated by `config_to_program` '
                f'because it does not accept any argument.')
            raise RuntimeError(msg)
        elif len(expected_args) > 1:
            msg = (f'Function `{fn.__name__}` cannot be decorated by `config_to_program` '
                f'because it expects {len(expected_args)} > 1; If you want to pass '
                f'extra arguments to this function, use kwargs with default values.')
            raise RuntimeError(msg)

    @classmethod
    def _check_args(cls, fn, args):
        if len(args):
            msg = (f'After decorating `{fn.__name__}` with `config_to_program`, '
                   f'do not provide any additional arguments while invoking it; '
                   f'any additional parameter should be passed as a kwarg.')
            raise RuntimeError(msg)

    @classmethod
    def destination_formatter(cls, param_name: str) -> str:
        param_name = param_name.replace('.', '__')
        param_name = f"__{param_name}"
        return param_name

    @classmethod
    def update_config(cls, config: dict, param_name: str, param_value: str) -> dict:
        config_copy = copy.deepcopy(config)

        param_name, *param_rest = param_name.split('.', 1)

        if len(param_rest) > 0:
            sub_config = config_copy.setdefault(param_name, {})
            param_value = cls.update_config(
                config=sub_config if isinstance(sub_config, dict) else {},
                param_name='.'.join(param_rest),
                param_value=param_value
            )

        config_copy[param_name] = param_value
        return config_copy

    def __new__(cls, config_node: Type[ConfigNode]) -> Callable:
        if not(isclass(config_node) or issubclass(config_node, ConfigNode)):
            raise ValueError(f'`config_node` is not a subclass of {ConfigNode.__name__}')

        local_config_node = config_node

        def main_decorator_wrapper(fn):
            cls._check_signature(fn)

            @functools.wraps(fn)
            def main_wrapped(*args, **kwargs):
                cls._check_args(fn=fn, args=args)

                ap = argparse.ArgumentParser(
                    prog=f'Parser for configuration {config_node.__name__}',
                    usage='',
                    )

                ap.add_argument('-c', dest='config', default=None, help='YAML config file')
                ap.add_argument('-p', dest='print', default=None, help='Print configuration',
                                choices=['input', 'i',
                                         'parsed', 'p',
                                         'continue', 'c'])

                all_params = ConfigNodeProps.get_all_parameters(local_config_node)

                for param_spec in all_params:
                    ap.add_argument(f'--{param_spec.name}',
                                    dest=cls.destination_formatter(param_spec.name),
                                    metavar='',
                                    type=type_evaluator(param_spec.type),
                                    default=argparse.SUPPRESS,
                                    help=(f"Type: {param_spec.type.__name__}; "
                                        f"Default: {param_spec.default}"))

                parsed_args = ap.parse_args()

                if parsed_args.config:
                    config = yaml.safe_load(read_raw_file(parsed_args.config))
                else:
                    config = {}

                for param_spec in all_params:
                    param_value = getattr(parsed_args,
                                          cls.destination_formatter(param_spec.name),
                                          argparse.SUPPRESS)
                    if param_value != argparse.SUPPRESS:
                        config = cls.update_config(config=config,
                                                   param_name=param_spec.name,
                                                   param_value=param_value)

                if parsed_args.print in {'input', 'i'}:
                    print('INPUT CONFIG:')
                    print(yaml.safe_dump(config, indent=2).strip())
                    return lambda: None

                config = local_config_node(config)

                if parsed_args.print in {'parsed', 'p', 'continue', 'c'}:
                    print('PARSED CONFIG:')
                    print(config_to_yaml(config, indent=2).strip())
                    if parsed_args.print not in {'continue', 'c'}:
                        return lambda: None

                return fn(config, **kwargs)

            return main_wrapped

        return main_decorator_wrapper


import logging
from multiprocessing import current_process
import os
from typing import Any, Callable, Dict, Type, TypeVar, Union


from .utils import mkdir_p


def configure_logging(
    file_logging_path: str = None,
    make_dir_if_missing: bool = True,
    fmt: str = "[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
    datefmt: str = "%Y-%m-%d %H:%M:%S",
    logging_level: int = logging.INFO,
    force_root_reattach: bool = True,
    root_formatter_kwargs: dict = None,
    stream_handler_kwargs: dict = None,
    file_handler_kwargs: dict = None,
    basic_config_kwargs: dict = None,
    additional_handlers: list = None,
):
    """A big function that speeds up configuration of logging"""

    # change how the formatter looks
    kw = root_formatter_kwargs or {}
    root_formatter = logging.Formatter(fmt=fmt, datefmt=datefmt, **kw)

    handlers = list(additional_handlers or [])

    kw = stream_handler_kwargs or {}
    root_stream_handler = logging.StreamHandler(**kw)
    root_stream_handler.setFormatter(root_formatter)
    handlers.append(root_stream_handler)

    if file_logging_path is not None:
        path_dir, path_name = os.path.split(file_logging_path)

        path_dir = mkdir_p(path_dir) if make_dir_if_missing else path_dir

        path_name, path_ext = os.path.splitext(path_name)
        path_name = f'{path_name}_{current_process().name}{path_ext}'

        kw = file_handler_kwargs or {}
        root_file_handler = logging.FileHandler(filename=os.path.join(path_dir, path_name), **kw)
        root_file_handler.setFormatter(root_formatter)
        handlers.append(root_file_handler)

    # add the handlers to the root logger, force reattaching other loggers
    kw = basic_config_kwargs or {}
    logging.basicConfig(level=logging_level, force=force_root_reattach, handlers=handlers, **kw)



from .node import (
    ConfigNode,
    ConfigFlexNode,
    ConfigParam
)
from .registry import ConfigRegistry
from .functional import (
    config_from_file,
    config_from_string,
    config_to_dict,
    config_to_yaml,
    config_to_program
)
from . import defaults as _defaults
from .instantiate import instantitate
from .logging import configure_logging




import errno
import os
from ast import literal_eval
from typing import Any, Callable, Type

import smart_open


def mkdir_p(path: str) -> str:
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno != errno.EEXIST or not os.path.isdir(path):
            raise
    return path

class hybridmethod:
    """A decorator that allows overloading a method depending
    on whether it is a class method or instance method. From
    https://stackoverflow.com/a/28238047"""

    def __init__(self, fclass, finstance=None, doc=None):
        self.fclass = fclass
        self.finstance = finstance
        self.__doc__ = doc or fclass.__doc__
        # support use on abstract base classes
        self.__isabstractmethod__ = bool(
            getattr(fclass, '__isabstractmethod__', False)
        )

    def classmethod(self, fclass):
        return type(self)(fclass, self.finstance, None)

    def instancemethod(self, finstance):
        return type(self)(self.fclass, finstance, self.__doc__)

    def __get__(self, instance, cls):
        if instance is None or self.finstance is None:
              # either bound to the class, or no instance method available
            return self.fclass.__get__(cls, None)
        return self.finstance.__get__(instance, cls)


class MISSING:
    """Used to keep track of missing parameters"""

    def __repr__(self) -> str:
        return 'MISSING'


def type_evaluator(field_type: Type[Any]) -> Callable:
    """Constructs a simple eval function for command line args
    that does proper casting depending on field_type"""

    def _type_fn(value: str) -> field_type:
        if not issubclass(field_type, str):
            # unless the type is string, we use literal
            # eval to cast to a built in python type;
            # literal_eval supports things such as list, dict, etc. too!
            value = literal_eval(value)

        return field_type(value)
    return _type_fn


def read_raw_file(file_path: str) -> str:
    with smart_open.open(file_path, mode='r', encoding='utf-8') as f:
        content = f.read()
    return content



import copy
import importlib
import functools
from typing import Any, Callable, Dict, Type, TypeVar, Union

from .node import ConfigNodeProps, ConfigNode
from .functional import config_from_dict


class InitLater(functools.partial):
    def get_kw(self, *args, **kwargs):
        return self.keywords.get(*args, **kwargs)

    def __bool__(self) -> bool:
        return len(self.keywords) > 0

    def __call__(self, /, *args, **keywords):
        # recursively call deferred initialization if
        # we encounter another InitLater
        keywords = {k: v() if isinstance(v, InitLater) else v
                    for k, v in {**self.keywords, **keywords}.items()}
        return self.func(*self.args, *args, **keywords)


class get_callable:
    @staticmethod
    def is_module(path):
        try:
            spec = importlib.util.find_spec(path)
            return spec is not None
        except ModuleNotFoundError:
            return False

    @classmethod
    def _get_callable(cls, path):
        if cls.is_module(path):
            return importlib.import_module(path)
        elif '.' in path:
            m_name, c_name = path.rsplit('.', 1)
            if cls.is_module(m_name):
                container = importlib.import_module(m_name)
            else:
                container = get_callable(m_name)
            return getattr(container, c_name)
        else:
            return globals().get(path, __builtins__.get(path, None))

    def __new__(cls, path):
        cl_ = cls._get_callable(path)

        if cl_ is None:
            raise ModuleNotFoundError(f'Could not find `{path}`')
        return cl_


class instantitate:
    TARGET = '_target_'

    @staticmethod
    def _no_op():
        ...

    @classmethod
    def later(
        cls,
        config: Union[Dict[str, Any], ConfigNode, None] = None,
        _recursive_: bool = True,
        **kwargs
    ) -> InitLater:
        """For now we use hydra instead of doing the lookup ourselves;
        we will fix it later"""
        if config is None:
            return InitLater(cls._no_op)

        config = (config_from_dict(config, flex=True)
                  if isinstance(config, dict) else
                  copy.deepcopy(config))

        if len(kwargs) > 0:
            config = config << config_from_dict(kwargs, flex=True)

        if cls.TARGET not in config:
            msg = f'Cannot instantiate from `{config}`: `{cls.TARGET}` keyword missing'
            raise ValueError(msg)

        fn = get_callable(ConfigNodeProps.get_props(config).pop(cls.TARGET))

        def _recursive_init(param):
            if (_recursive_ and
                isinstance(param, (ConfigNode, dict))
                and cls.TARGET in param):
                param = cls.later(config=param, _recursive_=True)
            return param

        init_call_dict = {k: _recursive_init(v) for k, v in config}

        return InitLater(fn, **init_call_dict)

    @classmethod
    def now(
        cls,
        config: Union[ConfigNode, dict] = None,
        _recursive_: bool = True,
        **kwargs
    ) -> object:
        """Create a later, but initialize it now!"""

        if config is None:
            return None

        return cls.later(config=config, _recursive_=_recursive_, **kwargs)()

    def __new__(cls, *args, **kwargs) -> object:
        """Alias for `instantitate.now`"""
        return cls.now(*args, **kwargs)
